#!/usr/bin/env python3
"""
Jewels

Usage:
	jewels interactive <field_file> [--timelimit=<m:s>] [--no-help]
	jewels automatic
"""

import sys
import datetime
import enum
from collections import namedtuple
import numpy
import docopt

class Game:
	"""Warning: self.field is row-major!"""
	Move = namedtuple("Move", ("from_pos", "to_pos", "gained_pts"))
	#group - tuple of members' positions

	class Type(enum.Enum):
		interactive = 0
		automatic = 1

	def __init__(self, inlines, gametype, timelimit=None):
		self.gametype = gametype
		self.timelimit = timelimit

		#for i, line in enumerate(inlines):
		inlines = list(map(lambda line: line[:-1], inlines)) #Strip newlines
		if self.gametype == Game.Type.interactive:
			self.field = self.parse_field(inlines[:8])
			self.available_stones = self.parse_available_stones(inlines[8:])
			self.available_stones_pos = [-1 for i in range(8)]
		else:
			assert(self.gametype == Game.Type.automatic)
			self.max_moves = int(inlines[0])
			self.field = self.parse_field(inlines[1:9])

	def parse_field(self, inlines):
		field = numpy.zeros((8, 8))
		for i, line in enumerate(inlines):
			for j, stone in enumerate(line):
				field[i, j] = int(stone)
		return field

	def parse_available_stones(self, inlines):
		available_stones = [[] for i in range(8)]
		for i, line in enumerate(inlines):
			available_stones[i - 8] += map(int, line)
		return available_stones

	def pos_out_of_bounds(self, pos):
		return pos[0] < 0 or pos[0] >= self.field.shape[0] or\
			pos[1] < 0 or pos[1] >= self.field.shape[1]

	def find_all_groups(self, field):
		found_groups = []
		for y in range(field.shape[0]):
			for x in range(field.shape[1]):
				stone = field[y, x]
				if stone == 0:
					continue
				for field_member_positions in (\
					((y + 1, x), (y + 2, x)),
					((y - 1, x), (y + 1, x)),
					((y - 2, x), (y - 1, x)),
					((y, x + 1), (y, x + 2)),
					((y, x - 1), (y, x + 1)),
					((y, x - 2), (y, x - 1))):
					if any(filter(self.pos_out_of_bounds,
						field_member_positions)):
						#A position in these fmps is out of bounds
						continue

					if field[field_member_positions[0]] == stone and\
						field[field_member_positions[1]] == stone:
						#Group found
						group = set((field_member_positions[0],
								field_member_positions[1],
								(y, x)))
						if group in found_groups:
							continue
						else:
							found_groups.append(group)
							yield group

	def cleanup_field(self, field, new_stones=True, no_disallowed=True):
		"""Warning! Modifies field unlike other functions"""
		groups = list(self.find_all_groups(field))
		if len(groups) == 0:
			return False if no_disallowed else True
		for group in groups:
			for pos in group:
				field[pos] = 0
		if new_stones:
			for y in range(field.shape[0]):
				for x in range(field.shape[1]):
					if field[y, x] == 0:
						self.available_stones_pos[y] =\
							self.available_stones_pos[y] + 1 if\
							self.available_stones_pos[y] + 1 <\
							len(self.available_stones[y]) else 0
						field[y, x] = self.available_stones[y][\
							self.available_stones_pos[y]]
		self.cleanup_field(field, new_stones=True,
			no_disallowed=False)


	def play_move(self, field, move, cleanup=True, new_stones=True,
			no_disallowed=True):
		res_field = field.copy()
		stone = field[move.from_pos]
		displaced_stone = field[move.to_pos]
		res_field = field.copy()
		res_field[move.to_pos] = stone
		res_field[move.from_pos] = displaced_stone
		if cleanup or new_stones or no_disallowed:
			retval = self.cleanup_field(res_field,
				no_disallowed=no_disallowed, new_stones=new_stones)
			if retval is False:
				return None

		return res_field

	def get_possible_moves(self, field):
		for y in range(field.shape[0]):
			for x in range(field.shape[1]):
				pos = (y, x)
				stone = field[pos]
				for direction in ((0, 1), (1, 0), (0, -1), (-1, 0)):
					npos = (y + direction[0], x + direction[1])
					if self.pos_out_of_bounds(npos):
						continue
					displaced_stone = field[npos]
					field_after_move = field.copy()
					field_after_move[npos] = stone
					field_after_move[pos] = displaced_stone
					groups = list(self.find_all_groups(field_after_move))
					if len(groups) > 0:
						yield Game.Move(pos, npos, len(groups) * 3)

#	def get_move_pts(self, field, move, pts=0, i=0):
#		if i >= 3:
#			return pts
#		res_field = self.play_move(field, move)
#		move_pts = []
#		for move in self.get_possible_moves(field):
#			move_pts.append((self.get_move_pts(
#				field, move, pts + move.gained_pts, i + 1)))
#		if len(move_pts) == 0:
#			return pts
#		return sorted(move_pts, reverse=True)[0]
#
#	def get_best_move_dfs(self, field, i=1):
#		best_moves = []
#		for move in list(self.get_possible_moves(field)):
#			print(self.get_move_pts(field, move))
#			best_moves.append(
#				self.get_best_move_dfs(self.play_move(field, move), i + 1))
#		return sorted(best_moves, key=lambda m: m.gained_pts, reverse=True)[0]

if __name__ == "__main__":
	args = docopt.docopt(__doc__)
	if args["interactive"] == True:
		if args["--timelimit"] is not None:
			try:
				timelimit_s= args["--timelimit"].split(":")
				minutes = int(timelimit_s[0])
				seconds = int(timelimit_s[1])
				timelimit = datetime.timedelta(minutes=minutes,
					seconds=seconds)
				print(timelimit)
			except (ValueError, IndexError):
				print("Time limit format: minutes:seconds")
				sys.exit(1)
		else:
			timelimit = None
		game = Game(open(args["<field_file>"]), Game.Type.interactive)
		import gui
		gui.run(game, timelimit, not args["--no-help"])
	else:
		assert(args["automatic"] == True)
		game = Game(open("vstup.txt"), Game.Type.automatic)
		outfile = open("vystup.txt", "w")
		moves = game.get_possible_moves(game.field)
		best_move = sorted(moves, key=lambda m: m.gained_pts, reverse=True)[0]
		outfile.write("{} {} {} {}".format(
			best_move.from_pos[1] + 1, best_move.from_pos[0] + 1,
			best_move.to_pos[1] + 1, best_move.to_pos[0] + 1))

