import sys
import math
from PyQt5.QtCore import *
from PyQt5.QtGui import *
from PyQt5.QtWidgets import *
import game

stone_colours = {
	0: QColor(0, 0, 0),
	1: QColor(255, 0, 0),
	2: QColor(0, 255, 0),
	3: QColor(0, 0, 255),
	4: QColor(255, 255, 255),
	5: QColor(255, 255, 0),
	6: QColor(255, 0, 255)
}

class MainWidget(QWidget):
	def __init__(self, game, timelimit, do_help, *args, **kwargs):
		super().__init__(*args, **kwargs)
		self.game = game
		self.timelimit = timelimit
		self.do_help = do_help
		self.end_timer = None
		self.game_over = False

		self.stone_width = 50
		self.stone_height = 50
		self.stone_padding = 10
		self.stone_border_colour = QColor(0, 0, 0)
		self.stone_selected_border_colour = QColor(127, 127, 127)
		self.stone_border_thickness = 2
		self.stone_selected_border_thickness = 5
		self.board_padding = 20
		self.header_padding = 20
		self.hint_colour = QColor(0, 0, 0)

		self.selected_stone = None
	
	def on_endtimer(self):
		self.end_timer.stop()
		self.game_over = True
		print("Game over!")

	def start_timelimit_timer(self):
		if self.timelimit is None:
			return
		self.end_timer = QTimer(self)
		self.end_timer.timeout.connect(self.on_endtimer)
		self.end_timer.start(self.timelimit.seconds * 1000)

	def mousePressEvent(self, e):
		if self.game_over:
			return

		field_x = math.floor((e.x() - self.board_padding) / (
			self.stone_padding + self.stone_width))
		field_y = math.floor((e.y() - (self.board_padding +
			self.header_padding * 2 + self.stone_height)) / (
			self.stone_padding + self.stone_height))
		stone_pos = (field_y, field_x)
		if self.game.pos_out_of_bounds(stone_pos):
			return
		if self.selected_stone is None:
			self.selected_stone = stone_pos
			if self.game.field[stone_pos] == 0:
				return
		elif abs(self.selected_stone[0] - stone_pos[0]) +\
			abs(self.selected_stone[1] - stone_pos[1]) == 1:
			move = game.Game.Move(self.selected_stone, stone_pos, None)
			res_field = self.game.play_move(self.game.field, move,
				no_disallowed=True)
			if res_field is None:
				return
			self.game.field = res_field
			self.selected_stone = None

			if len(list(self.game.get_possible_moves(self.game.field))) == 0:
				self.game_over = True
				print("Game over!")

		else:
			self.selected_stone = None

		self.repaint()
	
	def stone_pos_to_win_pos(self, stone_pos):
		return (
			stone_pos[1] * (self.stone_width + self.stone_padding) +\
			self.board_padding,
			
			stone_pos[0] * (self.stone_height + self.stone_padding) +\
			self.board_padding +\
			self.header_padding * 2 + self.stone_height
		)

	def draw_field(self, painter, field):
		painter.setPen(QPen(self.stone_border_colour,
			self.stone_border_thickness))
		for i, line in enumerate(self.game.available_stones):
			stone = line[self.game.available_stones_pos[i] + 1]
			xoffset = self.header_padding +\
				i * (self.stone_width + self.stone_padding)
			yoffset = self.header_padding
			colour = stone_colours[stone]
			painter.setBrush(QBrush(colour, Qt.SolidPattern))
			painter.drawEllipse(xoffset, yoffset,
				self.stone_width, self.stone_height)


		for y in range(field.shape[0]):
			for x in range(field.shape[1]):
				if (y, x) == self.selected_stone:
					painter.setPen(QPen(self.stone_selected_border_colour,
						self.stone_selected_border_thickness))

				xoffset, yoffset = self.stone_pos_to_win_pos((y, x))
				
				colour = stone_colours[field[y, x]]
				painter.setBrush(QBrush(colour, Qt.SolidPattern))
				painter.drawEllipse(xoffset, yoffset,
					self.stone_width, self.stone_height)
				if (y, x) == self.selected_stone:
						painter.setPen(QPen(self.stone_border_colour,
						self.stone_border_thickness))

		painter.setBrush(QBrush(self.hint_colour, Qt.SolidPattern))
		if self.do_help:
			for move in self.game.get_possible_moves(self.game.field):
				startx, starty = self.stone_pos_to_win_pos(move.from_pos)
				endx, endy =  self.stone_pos_to_win_pos(move.to_pos)
				painter.drawLine(
					startx + self.stone_width / 2,
					starty + self.stone_height / 2,
					endx + self.stone_width / 2,
					endy + self.stone_height / 2)

	def paintEvent(self, e):
		painter = QPainter(self)
		painter.setRenderHint(QPainter.Antialiasing, True)
		self.draw_field(painter, self.game.field)
		if self.end_timer is None:
			self.start_timelimit_timer()

def run(game, timelimit, do_help):
	app = QApplication(sys.argv)

	main_widget = MainWidget(game, timelimit, do_help)
	main_widget.show()

	app.exec()

